@@ -7,9 +7,12 @@ import json
 import math
 from time import strftime
 
+import fastremap
 import numpy as np
 from tqdm import tqdm
 
+from cloudfiles import CloudFiles
+
 import cloudvolume
 import cloudvolume.exceptions
 from cloudvolume import CloudVolume
@@ -21,8 +24,10 @@ from igneous.tasks import (
   ContrastNormalizationTask, DownsampleTask,
   TransferTask, TouchTask, LuminanceLevelsTask,
   HyperSquareConsensusTask, # HyperSquareTask,
-  ImageShardTransferTask, ImageShardDownsampleTask
+  ImageShardTransferTask, ImageShardDownsampleTask,
+  CCLFacesTask, CCLEquivalancesTask, RelabelCCLTask
 )
+
 from igneous.shards import image_shard_shape_from_spec
 from igneous.types import ShapeType
 
@@ -45,6 +50,9 @@ __all__  = [
   "create_hypersquare_consensus_tasks",
   "create_luminance_levels_tasks",
   "create_contrast_normalization_tasks",
+  "create_ccl_face_tasks",
+  "create_ccl_equivalence_tasks",
+  "create_ccl_relabel_tasks",
 ]
 
 # A reasonable size for processing large
@@ -1256,3 +1264,151 @@ def create_hypersquare_consensus_tasks(
         )
 
   return HyperSquareConsensusTaskIterator()
+
+
+def create_ccl_face_tasks(
+  cloudpath, mip, shape=(512,512,512),
+  threshold_gte:Optional[Union[float,int]] = None,
+  threshold_lte:Optional[Union[float,int]] = None,
+):
+  """pass 1"""
+  vol = CloudVolume(cloudpath, mip=mip)
+
+  shape = Vec(*shape)
+  bounds = vol.bounds.clone()    
+
+  class CCLFaceTaskIterator(FinelyDividedTaskIterator):
+    def task(self, shape, offset):
+      return partial(CCLFacesTask, 
+        cloudpath=cloudpath, 
+        mip=mip, 
+        shape=shape.clone(), 
+        offset=offset.clone(),
+        threshold_gte=threshold_gte,
+        threshold_lte=threshold_lte,
+      )
+
+    def on_finish(self):
+      vol.provenance.processing.append({
+        'method': {
+          'task': 'CCLFacesTask',
+          'cloudpath': cloudpath,
+          'mip': mip,
+          'shape': shape.tolist(),
+          'threshold_gte': threshold_gte,
+          'threshold_lte': threshold_lte,
+        },
+        'by': operator_contact(),
+        'date': strftime('%Y-%m-%d %H:%M %Z'),
+      })
+      vol.commit_provenance()
+
+  return CCLFaceTaskIterator(bounds, shape)
+
+def create_ccl_equivalence_tasks(
+  cloudpath, mip, shape=(512,512,512),
+  threshold_gte:Optional[Union[float,int]] = None,
+  threshold_lte:Optional[Union[float,int]] = None,
+):
+  """pass 2. Note: shape MUST match pass 1."""
+  vol = CloudVolume(cloudpath, mip=mip)
+
+  shape = Vec(*shape)
+  bounds = vol.bounds.clone()    
+
+  class CCLEquivalencesTaskIterator(FinelyDividedTaskIterator):
+    def task(self, shape, offset):
+      return partial(CCLEquivalancesTask, 
+        cloudpath=cloudpath, 
+        mip=mip, 
+        shape=shape.clone(), 
+        offset=offset.clone(),
+        threshold_gte=threshold_gte,
+        threshold_lte=threshold_lte,
+      )
+
+    def on_finish(self):
+      vol.provenance.processing.append({
+        'method': {
+          'task': 'CCLEquivalancesTask',
+          'cloudpath': cloudpath,
+          'mip': mip,
+          'shape': shape.tolist(),
+          'threshold_gte': threshold_gte,
+          'threshold_lte': threshold_lte,
+        },
+        'by': operator_contact(),
+        'date': strftime('%Y-%m-%d %H:%M %Z'),
+      })
+      vol.commit_provenance()
+
+  return CCLEquivalencesTaskIterator(bounds, shape)
+
+def create_ccl_relabel_tasks(
+  src_path, dest_path, 
+  mip, shape=(512,512,512),
+  chunk_size=None, encoding=None,
+  threshold_gte:Optional[Union[float,int]] = None,
+  threshold_lte:Optional[Union[float,int]] = None,
+):
+  """pass 3"""
+
+  src_vol = CloudVolume(src_path, mip=mip)
+
+  cf = CloudFiles(src_path)
+  max_label = int(cf.get_json(cf.join(src_vol.key, "ccl", "max_label.json"))[0])
+
+  smallest_dtype = fastremap.fit_dtype(np.uint64, max_label)
+  smallest_dtype = np.dtype(smallest_dtype).name
+
+  try:
+    dest_vol = CloudVolume(dest_path, mip=mip)
+  except cloudvolume.exceptions.InfoUnavailableError:
+    info = copy.deepcopy(src_vol.info)
+    info["data_type"] = smallest_dtype
+    scale = info["scales"][mip]
+    if chunk_size:
+      scale["chunk_sizes"] = [chunk_size]
+    if encoding:
+      scale["encoding"] = encoding
+      if encoding == "compressed_segmentation":
+        scale['compressed_segmentation_block_size'] = (8,8,8)
+    if "sharding" in scale:
+      del scale["sharding"]
+    dest_vol = CloudVolume(dest_path, info=info, mip=mip)
+    dest_vol.reset_scales()
+    dest_vol.commit_info()
+
+  shape = Vec(*shape)
+  bounds = src_vol.bounds.clone()  
+
+  class RelabelCCLTaskIterator(FinelyDividedTaskIterator):
+    def task(self, shape, offset):
+      return partial(RelabelCCLTask, 
+        src_path=src_path, 
+        dest_path=dest_path,
+        mip=mip, 
+        shape=shape.clone(), 
+        offset=offset.clone(),
+        threshold_gte=threshold_gte,
+        threshold_lte=threshold_lte,
+      )
+
+    def on_finish(self):
+      dest_vol.provenance.processing.append({
+        'method': {
+          'task': 'RelabelCCLTask',
+          'src_path': src_path, 
+          'dest_path': dest_path,
+          'mip': mip,
+          'shape': shape.tolist(),
+          'threshold_gte': threshold_gte,
+          'threshold_lte': threshold_lte,
+        },
+        'by': operator_contact(),
+        'date': strftime('%Y-%m-%d %H:%M %Z'),
+      })
+      dest_vol.commit_provenance()
+
+  return RelabelCCLTaskIterator(bounds, shape)
+
